.nolist
.include "include\m32def.inc"
.include "include\macros.inc" ; ADDI R16,3
.list

; Belegung der I/O-Ports
; ----------------------
;
; PortB3: (MOSI)
; PortB4: (MISO)								
; PortB5: (SCK)
;		Testbord
; PortD0: 	Rx
; PortD1: 	Tx
; PortD2:	Taster		
; PortD3: 	Taster		
; PortD4: 	Taster		
; PortD5: 	LED1	
; PortD6: 	LED2	
; PortD7:	Summer
;
; ------------------
; Belegung der Timer
; ------------------
;
; Timer0: ZMD    125s  
; Timer1: Interrupt,  Warteprogramm
; Timer2: Funkuhr 8 ms
;
; -----------------------------------
; Konstanten und Registerdefinitionen
; -----------------------------------
;
; Festwerte definieren
;
.equ	clock	= 8000000	;Controller Taktfrequenz= 8,0MHz
.equ	baud 	= 38400	;RS-232 Baudrate= 9600     38400
.equ 	UBRRVAL = CLOCK/(BAUD*16)-1
.equ	ZMD= PortC
.equ	Pinx = 2
;
; REGISTER definieren
;
.def	temp   	=R16	; allg. temporres Register
.def 	temp2 	=R17	; temp. register
.def 	flags	=r23    ; diverse Flags zur Steuerung
	  .equ neusek     =7 ; neue Sekunde eingetroffen
	  .equ Dallasflag =6 ; Dallas Temperaturmessung in Bearbeitung
	  .equ Relais     =5 ; Relais gesetzt	
	  .equ dcf_min_end=4 ; DCF_Uhrzeit anzeigen bei voller Minute	

;registers for DCF77 telegram
.def	dcf1 = R4
.def	dcf2 = R5
.def	dcf3 = R6
.def	dcf4 = R7
.def	dcf5 = R8

;DCF77 receiver
.equ	DCFPort = PORTD
.equ	DCFPin  = 2
.equ	DCFInvert = 0 ;inverted output (negative pulses), most receivers do this!
.equ	DCFPullup = 1 ;internal pull-up resistor for DCF77 input

;time/date registers
.def	subsec= R22
 
;DCF77-specific registers
.def	dcfbits = R18 ;received bits
.def	dcftime = R19 ;1/125-seconds since last edge
.def	flags1  = R20
.def	flags2  = R21

;bit definitions for the flags1 register
.equ	ErrBitCount  = 0 ;last error: wrong number of bits
.equ	ErrParity    = 1 ;last error: parity error
.equ	ErrBitLength = 2 ;last error: wrong bit length
.equ	ErrStartBit  = 3 ;last error: start bit was not 1
.equ	TimeOK       = 4 ;time is OK
.equ	DateOK       = 5 ;date is OK (reset at midnight)
.equ	StopDCF      = 6 ;stop DCF77 reception until a new telegram starts
.equ	DCFdisabled  = 7 ;DCF77 receiver disabled (reset when a new hour begins)
.equ	ErrMask = 1<<ErrBitLength|1<<ErrParity|1<<ErrBitCount|1<<ErrStartBit ;used to reset errors

;bit definitions for the flags2 register
.equ	LastState = DCFPin ;last state of DCF77 input, must be the same bit as on the port!

;==============
; SRAM Variablen
;==============
.dseg
.org	SRAM_START
;
Schreibsendezeig:.byte   1 ; Schreib   Sendezeiger
Lesensendezeig:	 .byte   1 ; Lesen     Sendezeiger
Schreibempfzeig: .byte   1 ; Schreiben Empfangszeiger
Lesenempfzeig:	 .byte   1 ; Lesen     Empfangszeiger
Uhr_bcd:
;Subcount: 	 .byte   1 ; 1/100 sek. 
Sekunden_bcd:	 .byte   1 ; Sekunden-einer
Minuten_bcd:	 .byte   1 ; Minuten-einer
Stunden_bcd:	 .byte   1 ; Stunden-einer
Tag_bcd:	 .byte   1 ; Tag
Monat_bcd:	 .byte   1 ; Monat
Jahr_bcd:	 .byte   1 ; Jahr
Wochentag_bcd:	 .byte   1 ; Wochentag (1=Montag 7=Sonntag)
Uhrend:		 .byte   0 ; dummy
ZMDPort1:	 .byte	 1 ; ZMD Port 1
ZMDPin1: 	 .byte   1 ; ZMD Pin  1
ZMDPort2:	 .byte	 1 ; ZMD Port 2
ZMDPin2: 	 .byte   1 ; ZMD Pin  2
ZMDPort3:	 .byte	 1 ; ZMD Port 3
ZMDPin3: 	 .byte   1 ; ZMD Pin  3
ZMDPort4:	 .byte	 1 ; ZMD Port 4
ZMDPin4: 	 .byte   1 ; ZMD Pin  4
ZMDPort5:	 .byte	 1 ; ZMD Port 5
ZMDPin5: 	 .byte   1 ; ZMD Pin  5
Variablenend:	 .byte   0 ; dummy
.org 	0x200	
Sendepuff:	 .byte 256 ; RS232 Sendepuffer
Sendepuffend:	 .byte   0 ; dummy
.org	0x300
Empfpuff:	 .byte 256 ; RS232 Empfangspuffer
Empfpuffend:	 .byte   0 ; dummy
.org	0x400	

; ===============
; Programmbereich
; ===============
.cseg			;Beginn eines Code-Segmentes
; Interrupt-Vektoren fr ATmega 32						  
;
.org	0		rjmp start	;Programm-Start
.org	INT0addr  	reti		;External Interrupt Request 0
.org	INT1addr	reti		;External Interrupt Request 1
.org	INT2addr	reti		;External Interrupt Request 2      nicht in ATMega 8 
.org	OC2addr 	rjmp oc2int	;Timer/Counter2 Compare Match 8ms, Funkuhr
.org	OVF2addr	reti		;Timer/Counter2 Overflow
.org	ICP1addr	reti		;Timer/Counter1 Capture Event
.org	OC1Aaddr	reti ;ocr1a	;Timer/Counter1 Compare Match A, frei
.org	OC1Baddr	reti ;ocr1b	;Timer/Counter1 Compare Match B, Warte
.org	OVF1addr	reti		;Timer/Counter1 Overflow
.org	OC0addr		reti ;oc0int	;Timer/Counter0 Compare Match, ZMD  nicht in ATMega 8
.org	OVF0addr	reti		;Timer/Counter0 Overflow
.org	SPIaddr		reti		;SPI, STC Serial Transfer Complete
.org	URXCaddr	rjmp UART_Rx_C	;RXC USART, Rx Complete
.org	UDREaddr	rjmp UART_Empty	;UDRE USART Data Register Empty
.org	UTXCaddr	reti ;UART_Tx_C	;TXC USART, Tx Complete
.org	ADCCaddr	reti	 	;ADC Conversion Complete
.org	ERDYaddr	reti		;EEPROM Ready
.org	ACIaddr		reti		;Analog Comparator
.org	TWIaddr		reti		;Two-wire Serial Interface
.org	SPMRaddr	reti		;Store Program Memory Ready
.org    INT_VECTORS_SIZE
Flashzaehler: .dw 1
START:
;      Stackpointer	
	ldi	temp,low (ramend)
	out	spl,temp
	ldi	temp,high(ramend)
	out	sph,temp		;Stackpointer setzen

;  ldi temp,0x63		;Oszillator Calibration Byte
;  out OSCCAL,temp		; Ausgabe
            
; Initialisierung SRAM, I/O-Ports, Hardware
;
;  RAM lschen
	ldi	zl,0x60
	ldi	zh,0			;Anfangsadresse SRAM
	clr	temp
sramcl:	st	z+,temp			;Speicherzelle lschen
	cpi	zl,low (ramend+1)	;letzte Speicherzelle erreicht?
	brne	sramcl			;nein -> Schleife
	cpi	zh,high(ramend+1)	;letzte Speicherzelle erreicht?
	brne	sramcl			;nein -> Schleife

; PORT D initialisieren	     ; Testbord	
	ldi temp, 0b11100010 ; 0,2-4=Eingnge; 1,5-7=Ausgnge 2-4=Taster 5,6=LED 7=Summer 0=Rx 1=Tx 
	out DDRD, temp
; PORT D setzen					
	ldi temp, 0b00011101 ; PULL-UPs (ca. 30 kOhm gegen +5V)	
	out PORTD, temp				

	ldi temp,$FF
	out portb,temp
	out porta,temp
	out portc,temp
;   DCF ports
	cbi	DCFPort-1, DCFPin 	;DDRx
	.if ( DCFPullup )
		sbi	DCFPort, DCFPin ;Pullup on
	.else	cbi	DCFPort, DCFPin ;Pullup off
	.endif

;    UART
	 ldi temp,  LOW (UBRRVAL)	;Baudrate einstellen
	 out UBRRL, temp
    	 ldi temp,  HIGH(UBRRVAL)
   	 out UBRRH, temp                ; auf 9600 Baud
 	ldi	temp,	 (1<<URSEL)| (0<<UMSEL)|(0<<UPM1)|(0<<UPM0)|(0<<USBS)|  (3<<UCSZ0)	
  	out	ucsrc,	temp       ;  asynchron ,kein Parity,       1 Stoppbit, 8 Datenbits
	ldi	temp,	     (1<<RXCIE) |      (1<<TXCIE)   |   (1<<UDRIE)| (1<<RXEN) | (1<<TXEN)
	out	UCSRB,temp ; Empfang komplett, Senden komplett, Senden leer, Empfang,     Senden 
;        Pufferzeiger initialisieren, stehen schon auf 00, der Seitengrenze

;  Timer1B			fr Wartezeit 5s bis 65 ms
	ldi temp, (0<< ctc1)|(1<<cs11)	; durchlaufender Zhler ,Vorteiler = 8
	out TCCR1B, temp
;ldi temp, high( clock/8000 *10 - 1 )	; Compare Wert A setzen auf 10ms
;out OCR1AH, temp
;ldi temp, low ( clock/8000 *10 - 1 ) 
;out OCR1AL, temp 

;  Timer0   48 khz, CTC, Vorteiler wird wegen ausreich. Genauigkeit nicht resettet. fr ZMD Sensor
	ldi temp, (1<< ctc0)|(1<<cs01)	; CTC-Modus, Vorteiler = 8
	out TCCR0,temp
	ldi temp, low( clock/8*31 - 1 ) ; Compare Wert setzen auf 31s  (clock/Vorteiler * s)
	out OCR0, temp
;  Timer2, ctc, Vorteiler/256, oc=250, 8 Mhz ---- 125 OV = 1 sek. 150=1,2 sek. DCF-Uhr
	ldi temp, (1<< ctc2)|(1<<cs21)|(1<<cs22); CTC-Modus, Vorteiler = 256, resetbar
	out TCCR2,temp
	ldi temp, ( 250 - 1 ) ; Compare Wert setzen auf 8ms  (clock/Vorteiler *32s)
	out OCR2, temp	
	ldi temp, (1<<OCIE0)|(1<<OCIE1B)|(1<<OCIE2); Compare Interrupt fr Timer 0,1B,2 aktivieren	
	out TIMSK,temp 
; TOIE1 = Overflow int. 65536s

; Registers
	ldi	dcfbits, 0
	ldi	dcftime, 0
	ldi	flags1, 1<<StopDCF
	ldi	flags2, 0
;   Register Z initialisieren
	ldi	zl,low (2*ruhe)		; Anfangseinsprung initialisieren fr ZMD Sensor
	ldi	zh,high(2*ruhe)
	ldi	temp,8
	mov	r15,temp
	clr 	flags			; Flags lschen

;disable analog comparator
;	ldi	temp, 0x80
;	out	ACSR, temp

;   Sleep
	ldi temp,1<<se          	; Sleep im Mode Idle
 	out mcucr,temp                  ; vorbereiten

	sei				; globale Interrupts einschalten
ldi    ZH,0
ldi    ZL,PORTC+0x20
ld     r16,Z
ldi	r16,pinc+0x20
in r16,pinc
;     Initialisierung Ende  -----------------------------------------------------------------------------

; =====================================================================
; Hauptprogrammschleife
; =====================================================================
;
; Liste der UART-Steuercodes empfangen:
; R     Relais umschalten (im Test LED)
; EW	EEPROM speichern
; ER	EEPROM lesen
; Liste der UART-Steuercodes senden:
; Z	Uhrzeit	
; T1	Temperatur ZMD
; T2	Temperatur Platin
; T3 	Temperatur Dallas
; T4	Temperatur Sirion
; F	Feuchte	   Sirion	
; I1    Strom Phase 1
; I2    Strom Phase 2
; I3    Strom Phase 3
; U	Spannung 230V
main:	
	sbrs flags,neusek         ;neue Sekunde eingetroffen? - ja
	rjmp next1		  ; nein---nchste Aufgabe
;DCF77 receiver 
	sbrs	flags,dcf_min_end ; volle DCF-Minute? - ja
	rjmp 	Zeit		  ; nein
dcf_eval:			  ; evaluate DCF77 telegram
    ;check date parity
	ldi	temp2, 0
	mov	temp, dcf3
	andi	temp, 0xFE
	rcall	parity
	mov	temp, dcf4
	rcall	parity
	mov	temp, dcf5
	andi	temp, 0x7F
	rcall	parity
	mov	temp, dcf5
	lsl	temp ;Bit7 -> C
	rol	temp ;C -> Bit0
	andi	temp, 0x01
	cp	temp, temp2
	brne	dcf_parityerror
    ;check hour parity
	ldi	temp2, 0
	mov	temp, dcf2
	andi	temp, 0xFC
	rcall	parity
	mov	temp, dcf3
	andi	temp, 0x01
	cp	temp, temp2
	brne	dcf_parityerror
    ;check minute parity
	ldi	temp2, 0
	mov	temp, dcf1
	andi	temp, 0xFC
	rcall	parity
	mov	temp, dcf2
	andi	temp, 0x01
	rcall	parity
	mov	temp, dcf2
	lsr	temp
	andi	temp, 0x01
	cp	temp, temp2
	brne	dcf_parityerror
   ;save minutes
	lsr	dcf2
	ror	dcf1
	lsr	dcf1
	sts	minuten_bcd, dcf1
   ;save hours
	lsr	dcf2
	sts	Stunden_bcd,dcf2
   ;save day
	lsl	dcf3 ;align day
	rol	dcf4 ;align month[5] weekday[3]
	rol	dcf5 ;align year
	lsr	dcf3
	lsr	dcf3
	sts	Tag_bcd,dcf3
   ;save weekday
	mov 	temp,dcf4
	andi	temp,0x07
	sts	Wochentag_bcd,temp
   ;save month
	lsr	dcf4
	lsr	dcf4
	lsr	dcf4
	sts	Monat_bcd,dcf4
   ;save year
	sts	Jahr_bcd,dcf5
   ;set time & date ok, disable DCF77 reception
	clr	subsec
	clr	temp
	sts	sekunden_bcd,temp
	ori	flags1, 1<<TimeOK | 1<<DateOK | 1<<DCFdisabled
	rjmp	Zeit
dcf_parityerror:
   	ori	flags1, 1<<ErrParity   ;set parity error
Zeit:
   ;increment seconds
	lds	temp,sekunden_bcd
	rcall	bcdinc
	sts	sekunden_bcd,temp
	cpi	temp, 0x60
	brlo	inctime_end
   ;increment minutes
	clr	temp
	sts	sekunden_bcd,temp
	lds	temp,minuten_bcd
	rcall	bcdinc
	sts	minuten_bcd,temp
	cpi	temp, 0x04 ;disable DCF77 reception if nothing received for 4 minutes
	brne	inctime_nodisable
	sbrc	flags1, TimeOK
	ori	flags1, 1<<DCFdisabled
  inctime_nodisable:
	cpi	temp, 0x60
	brlo	inctime_end
   ;increment hours
	clr	temp
	sts	minuten_bcd,temp
	andi	flags1, ~(1<<DCFdisabled) ;re-enable DCF77 reception
	lds	temp,Stunden_bcd
	rcall	bcdinc
	sts	Stunden_bcd,temp
	cpi	temp, 0x24
	brlo	inctime_end
   ;new day: mark date as bad until the new one was received
	clr	temp
	sts	stunden_bcd,temp
	andi	flags1, ~(1<<DateOK)
inctime_end:
;  Uhrzeit ausgeben
	ldi	temp,'Z'	; Kennzeichen Uhrzeit ausgeben
	rcall 	RSsend		;
	ldi	yh,high(Uhrend)
	ldi	yl,low (Uhrend)
	ldi	temp2,7		; 7 Speicherstellen=14 Ziffern ausgeben
    nexttime:
	ld	temp,-y
	lsr	temp
	lsr	temp
	lsr	temp
	lsr 	temp
	addi	temp,'0'	; BCD in Ascci umwandeln
	rcall	rssend
	ld	temp,y
	andi	temp,$0F
	addi	temp,'0'	; BCD in Ascci umwandeln
	rcall	rssend
	dec	temp2
	brne	nexttime	; nchste Speicherstelle 
	ldi	temp,10		; nchste Zeile
	rcall	rssend
	rcall 	togg6		; Toggle LED2
	cbr flags,1<<neusek  	;Jobflag lschen
;Zeichenuhr:
;	rcall	rssend 		; an RS232 senden
;	ret

next1:				; nchste Aufgabe

; UART Empfangen (RS232)
; zu empfangendes Zeichen  -> Register R16
RSEMPF:
		lds	xl,lesenempfzeig
		lds	xh,schreibempfzeig
		cp 	xl,xh
		breq	RSempfende		; Puffer leer?
		ldi	xh,high(empfpuff)
		ld	temp,x+			; nein, Zeichen in Temp
		sts	lesenempfzeig,xl
;    Interpretation
	cpi	temp,'E'
	breq	EEPROMspeich
	cpi	temp,'R'	; Relais		
	breq	RelaisNr
	sbrs	Flags,Relais	; Relais Flag =1?
	rjmp	RSempfende	; nein
	cpi	temp,'1'	; ja
	breq	Rel1
	cpi	temp,'2'
	breq	Rel2
	rjmp 	RSempfende
EEPROMspeich:
	rjmp 	RSempfende
RelaisNr: 
	sbr	flags,1<<Relais	; Flag Relais setzen
	rjmp 	RSempfende
Rel1:
	rcall 	togg5
	rjmp 	RSempfende1
Rel2:
	rcall 	togg7
	rjmp 	RSempfende1
RSempfende1:
	cbr	flags,1<<Relais	; Flag Relais lschen
RSempfende:	
next2:						; nchste Aufgabe
	
	cpi	flags,0
	brne	W6
	sleep             	     		;Rest der Runde pennen
   W6:	rjmp	 main
; Ende der Hauptprogrammschleife
;;############################################################################
 
;     Interrupts
;-----------------------------------------------------------------------------
;    ZMD Sensoren       5 I/O  Eingang	Port C 2-6
;  R15 Timingzhler =8
; R14  decrementzhler
; R25 Bitspeicher
; R12 Bitzhler
; R17 Bytespeicher
; R10 Bytezhler
; R9   Paritiy Addierer
;  R8 Sicherung von R15
;  Z= Sprungregister
; R13=SensorNr 1,2,4,8,16,32,64,128
.def SensorNr=R24
; .equ	 ZMD=PinC
		ldi 	SensorNr,1	
		ldi	temp, low( clock/8*31 - 1 )	; Compare Wert setzen auf 31s  ;(clock/Vorteiler * s) evtl. 47s fr Anschluerkennung mit einem Byte
		out	OCR2, temp
		ldi	temp,2				; 2 Bytes lesen
		mov	r10,temp  
OC0INT:						; Interrupt Timer2 Output Compare 48 khz
		push 	temp			; R16 sichern (temp) 2
		in	TEMP,SREG		; Einlesen des SREG  1
		push	TEMP			; Schreiben von  SREG  im Stack (KOPIE) 2
		in 	Temp,ZMD
		and	temp,SensorNr
		ijmp				; 2
Ruhe:		
		breq	return1			; 	; zurck, wenn low
		dec	r15				; 8x High testen
		brne	return				; noch keine 8 High-Pegel
						; gltige Ruhestellung
		ldi	zl,low (2*Startbitlow)
		ldi	zh,high(2*Startbitlow)
		rjmp	return
Startbitlow:	brne	return				; immer noch high=Ruhestellung
							; Low= Startbit erkannt
		ldi	temp,'T'			; Kennung T vor der Temp. setzen 
		rcall	rssend				; (in Ringspeicher abspeichern)
		ldi	zl,low (2*startbitmess)		; 
		ldi	zh,high(2*startbitmess)
		rjmp	return
Startbitmess:				; Genauigkeit 3 Takte
		sbis	PINC,2				; 1    +2
		rjmp	Startbitmess			; warten bis High 2
  Startbithigh:						; Highzeit messen, Genauigkeit 4 Takte
		inc	r15				; 1   R15=500 : 4= 125 bei 8 Mhz
		sbic	PINC,2				; 1    +2	
		rjmp 	Startbithigh			; 2
					 
		mov	temp,r15
		lsr	temp				; R15/2 
		out 	OCR2,temp			; in Compare  mit Vorteiler=8 (62s)
		ldi	Temp,PSR2
		out	SFIOR,temp			; und Vorteiler=0
		ldi	zl,low (2*bit)			; 1
		ldi	zh,high(2*bit)			; 1	
		rjmp	return				; 2
Bit:							; compare Zeit verdoppeln (124s)
		out 	OCR2,r15			; in Compare  mit Vorteiler=8
nextbyte:	ldi	zl,low (2*nextbit)		; nchstes Bit
		ldi	zh,high(2*nextbit)
		ldi	temp,8				; 8 Bit in Bitzhler 
		mov	r12,temp			; R12
nextBit:
		clc					; clear carry
		breq	W4
		sec					; set carry
	W4:	clr	temp
		adc	R9,temp				; Parity berechnen
		rol	r17				; carry nach Bit 0, Shift links
		dec 	r12				; Bitzhler 
		brne	return 				; =0?
		ldi	zl,low (2*Pariti)
		ldi	zh,high(2*Pariti)		
		rjmp	return
Pariti:
		ldi	r25,0				; clear carry
		breq	W5
		inc	r25				; set carry
	W5:	mov 	temp,r9
		andi	temp,01				; berechnetes Parity
		eor	r25,temp				; exor
		ldi	zl,low (2*stopbit)		; ja, zum halben Stopbit
		ldi	zh,high(2*stopbit)
		breq	return
 		ldi	zl,low (2*ruhe)			; bei Parityfehler von Vorne neu lesen
		ldi	zh,high(2*ruhe)		
		rjmp	return1
stopbit:
		mov 	temp,r17			; Byte in R17
		rcall 	rssend				; in RS232 Ringpuffer speichern
		ldi	zl,low (2*nextbyte)
		ldi	zh,high(2*nextbyte)		; zum nchsten Bit lesen
		dec	r10				; Bytezhler
		brne	return				; wenn 2 Bytes abgespeichert,
		ldi	zl,low (2*Sensor2)		;  ist das Lesen eines Sensors beendet. 
		ldi	zh,high(2*Sensor2)		; Startadresse zum nchsten lesen
  return1:
		ldi	temp,8
		mov 	r15,temp
  return:	
		pop	TEMP			; LESEN von SREG vom STACK (KOPIE) 2
		out	SREG,TEMP		; Wiederherstellen von SREG	   1 	
		pop	temp			; R16 (temp) zurck laden	   2
		reti				; zurck			   5 
Sensor2:	rjmp	ruhe

;	movw	r3:r2,x
;	movw	x,r3:r2	

zmd1:	in temp,pinc
	bst temp,2
	ret
zmd2: 	in temp,pinc
	bst temp,3
	ret

;*******************************************************************************
;*   DCF77 subroutines
;*******************************************************************************
;---------------------------------------
; Timer 2 output compare int (125 Hz)
;   modifies: dcftime, dcfbits, flags1, flags2, dcf1, dcf2, dcf3, dcf4, dcf5, subsec
;---------------------------------------
oc2int:
	push	temp
	ldi 	temp,1<<psr2	; Reset Vorteiler Timer2
	out	sfior,temp
	in	temp, SREG
	push	temp
	
	;don't evaluate DCF77 signal when reception is disabled
	sbrc	flags1, DCFdisabled
	rjmp	dcf_end
	
	inc	dcftime
   ;compare DCF77 input to last saved state
	in	temp, DCFPort-2 ;PINx
	eor	temp, flags2
	andi	temp, 1<<LastState
	.if ( DCFInvert ) 
			  .message "DCF77 Signal: INVERTED"
			   brne	dcf_end 		;no change
	.else		  .message "DCF77 Signal: normal"
			   breq	dcf_end 		;no change
	.endif
	
	;take appropriate action for pos/neg edges
	sbrs	flags2, LastState
	rjmp	dcf_posedge
	
   ; < negative edge ( 1 -> 0 ) >
	
	andi	flags2, ~(1<<LastState)
    ;exit if an error was detected before
	sbrc	flags1, StopDCF
	rjmp	dcf_end
	
	;increment bit number, exit if more than 59 bits were already received
	inc	dcfbits
	cpi	dcfbits, 60
	brsh	W3
dcf_evalbit:				;check bit length 
	cpi	dcftime, 10  ;bit is 0:  5 <= dcftime <= 7 ; 5= ( 78,125 ms)
	brlo	dcf_errbit
	cpi	dcftime, 16				   ; 8= (125     ms) 
	brlo	dcf_clearbit
	cpi	dcftime, 22 ;bit is 1: 11 <= dcftime <= 13 ;11= (171,875 ms)
	brlo	dcf_errbit
	cpi	dcftime, 27				   ;14= (218,75  ms)
	brlo	dcf_setbit
dcf_errbit:				;received bit is erroneous
	andi	flags1, ~ErrMask
	ori	flags1, 1<<ErrBitLength | 1<<StopDCF
	rjmp	dcf_end
	
	;save bit
dcf_clearbit:
	clc 		;clear bit
	rjmp	dcf_savebit
dcf_setbit:
	sec 		;set bit
dcf_savebit:
	ror	dcf5
	ror	dcf4
	ror	dcf3
	ror	dcf2
	ror	dcf1
	rjmp	dcf_end
    W3:	andi	flags1, ~ErrMask
	ori	flags1, 1<<ErrBitCount | 1<<StopDCF
	rjmp	dcf_end
	
   ; < positive edge ( 0 -> 1 ) >
dcf_posedge:
	ori	flags2, 1<<LastState
	
	;check if time was >1.2s (start of new telegram)
	cpi	dcftime, 150
	ldi	dcftime, 0
	brlo	dcf_end
	
	;end of telegram / new telegram
	andi	flags1, ~ErrMask & ~(1<<StopDCF)
	cpi	dcfbits, 59
	ldi	dcfbits, 0
	brne	W2
	sbr	flags,1<<dcf_min_end
	rjmp	dcf_end
    W2:	ori	flags1, 1<<ErrBitCount
	cbr	flags,1<<dcf_min_end
dcf_end:
;---------------------------------------
; increment time (fallback quartz clock)
;---------------------------------------
   ;increment 1/125-seconds (im Original=1/64 sek.)
	inc	subsec
	cpi	subsec,125
	brlo	subtime_end
	clr	subsec		
	sbr 	flags,1<<neusek        ;Job (Ausgabe) anmelden
subtime_end:
	pop	temp
	out	SREG, temp
	pop	temp
	reti

;*===============================
;*   GENERAL ARITHMETIC FUNCTIONS
;*===============================
;---------------------------------------
; calculate parity and add up
;   in:  temp
;   out: temp2 (add to existing value)
;---------------------------------------
parity:
	lsr	temp
	brcc	PC+2
	inc	temp2
	cpi	temp, 0
	brne	parity
	andi	temp2, 0x01
	ret
;---------------------------------------
; increment packed bcd number
;   in:  temp
;   out: temp
;---------------------------------------
bcdinc:
	subi	temp,-7 ;increment and add 6
	brhc	PC+2 ;half-carry is inverted when using "subi rh, -k" instead of add!
	subi	temp, 6
	ret
/*	
;---------------------------------------
; convert binary to packed BCD
;   in:       temp
;   out:      temp
;   modifies: temp2
;---------------------------------------
bin2bcd:
	ldi	temp2, 0
bin2bcd_loop:
	addi	temp2, 0x10
	subi	temp, 10
	brcc	bin2bcd_loop
	subi	temp2, 0x10
	addi	temp, 10
	add	temp, temp2
	ret			*/

;############################################################################
UART_Rx_C:		;RXC USART, Rx Complete
		push  	temp
		in	TEMP,SREG		; Einlesen des SREG 
		push	TEMP			; Schreiben von  SREG  im Stack (KOPIE)
		movw	r3:r2,x
;Empfangspuffer voll?
		in	temp,UDR		
		lds	xl,schreibempfzeig
		lds	xh,lesenempfzeig
		dec 	xh
		cp	xl,xh
		breq	Empfangende		; Puffer ist voll
		ldi	xh,high(empfpuff)
		st	x+,temp			; im Empfangspuffer abspeichern, Zeiger erhhen
		sts	schreibempfzeig ,xl	; und abspeichern
Empfangende:	
		movw	x,r3:r2		
		pop	TEMP			; LESEN von SREG vom STACK (KOPIE)
		out	SREG,TEMP		; Wiederherstellen von SREG
		pop  	temp			; R16 (temp) zurck laden	
		reti				; zurck

UART_Empty:		;UDRE USART Data Register Empty Interrupt
		push  	temp
		in	TEMP,SREG		; Einlesen des SREG 
		push	TEMP			; Schreiben von  SREG  im Stack (KOPIE)
		movw	r3:r2,x
;Sendepuffer leer?
		lds	xl,  lesensendezeig
		lds	xh,schreibsendezeig
		cp	xl,xh
		brne	senden			; nein- senden
		cbi	UCSRB,UDRIE		; ja - UDRE Interrupt aus 
		rjmp	sendeende
Senden:		ldi	xh, high(sendepuff)
		ld	temp,x+			; Byte aus Sendepuffer in UART, Zeiger erhhen
		out	UDR,temp
		sts	lesensendezeig ,xl
Sendeende:	
		movw	x,r3:r2	
		pop	TEMP			; LESEN von SREG vom STACK (KOPIE)
		out	SREG,TEMP		; Wiederherstellen von SREG
		pop  	temp			; R16 (temp) zurck laden	
		reti				; zurck

UART_Tx_C:		;TXC USART, Tx Complete
		reti
;---------------------------------------------------------------------------------------------------
; Unterprogramme (Subroutinen) allgemein
;
;           RS232 Ausgaben

RSSEND:			;Zeichen in Temp in UART Sendepuffer speichern
;Puffer voll?
		lds	xl, schreibsendezeig 
Sendewarten:	lds	xh, lesensendezeig
		dec	xh
		cp	xl,xh
		breq	Sendewarten		; ja - warten, noch voll
		ldi	xh,high(Sendepuff)	; nein
		st	x+,temp			; Zeichen abspeichern, Zeiger erhhen
		sts	schreibsendezeig ,xl
		sbi	UCSRB,UDRIE		; UDRE Interrupt ein
		ret

;====== einige Debug-Routinen ========

togg5:			;toggelt PD5
 push r17
 ldi r17,1<<5
 rjmp togg_weiter

togg6:                  ;toggelt PD6
 push r17                   ;Variable beschaffen
 ldi r17,1<<6               ;Toggelbit
togg_weiter:            ;Einsprunglabel
 push r16                   ;noch 'ne Variable beschaffen
 in r16,portd               ;Zustand einlesen
 eor r16,r17                ;Bit toggeln
 out portd,r16              ;und zurckschreiben
 pop r16                    ;Variablen 
 pop r17                    ;entsorgen
 ret                        ;fertig, zurck...

togg7:                  ;toggelt PD7
 push r17                   ;Variable beschaffen
 ldi r17,1<<7               ;Toggelbit
 rjmp togg_weiter           ;weiter wie gehabt...

/* Aus der Mainloop oder anderen Aufgaben ohne Wartezeit wird zum Programm mit
 Wartezeiten gesprungen. Beim 1. Mal mu das Rckkehrregister getrennt geladen
 werden. Von dort wird nach Ablauf der Teilaufgabe die Wartezeit gesetzt und 
 zur Main zurckgesprungen. Nach der Wartezeitende wird der Int. aufgerufen.
 In der Interruptroutine wird die Rckkehradresse vom Stack gesichert und die 
 Adresse der nchsten Teilaufgabe auf den Stack gelegt. Mit RETI wird dann zur
 nchsten Teilaufgabe verzweigt. Nach der Abarbeitung der Teilaufgabe wird die
 nchste Wartezeit gesetzt und zur gesicherten Rckkehradresse in der Main 
 gesprungen. */

;Mainloop1:	
	sbrc flags,Dallasflag   ;Wartezeit in Bearbeitung? 
	rjmp next		; ja---nchste Aufgabe   
	ldi temp ,low (next*2)	; nein				1
	ldi temp2,high(next*2)	;				1
	movw r12,temp		;				1
	 			; Aufgabe mit Wartezeiten	2
	sbr flags,1<<Dallasflag		
			; Teilaufgabe 1
	ldi temp,5	; Wartezeit in s, max.10000s, min 5s 1
	ldi temp2,0	;					1
	rcall warten	;					3
			; Teilaufgabe 2
	ldi temp,5	; Wartezeit in s, max.10000s, min 5s
	ldi temp2,0
	rcall warten
			; Teilaufgabe 3
	cbr flags,1<<Dallasflag
	rjmp	endaufg ; Dallas Ende		
next:

warten:			; warten, s in Temp2-Temp
	cli		; Interrupt sperren			1
	subi temp,4	; Laufzeit der Warteroutine abziehen,29Takte    1
	sbci r17,0	; 					1
	in  r10,TCNT1L	; 					1
	in  r11,TCNT1H	; Timer 1 lesen				1
	add r10,temp	; Wartezeit addieren			1
	adc r11,r17	;					1
	out OCR1BH, r11	; nein					1
	out OCR1BL, r10	; in OCR1B schreiben			1
	pop  r4		; Adresse der nchsten Teilaufgabe sichern, high       2
	pop  r5		;		" 		,low	2	
	in temp,TIMSK	; output compare int. 1B freigeben	1
	sbr temp,1<<OCIE1B ;					1
	out TIMSK,temp	;					1
	sei		; Interrupt freigeben			1			
endaufg:movw z,r12	; gesicherte Rckkehradresse nach Z	1
	ijmp		; zurck zur Main oder verlassenes Programm durch Int. 2
	
ocr1b:			; Compare Interrupt von Timer 1B
	push Temp
	in temp,TIMSK	;					1	
	cbr temp,1<<OCIE1B ;					1
	out TIMSK,temp	; output compare int. 1B sperren	1
	pop Temp
	pop r13		; Rckkehradresse sichern,high		2
	pop r12		; Rckkehradresse sichern,low		2
	push r5		; Adresse der nchsten Teilaufgabe auf Stack,low  2
	push r4		;		"		,high	2
	reti		;					4
.exit			; Ende des Programms
;===================================================================================================
set porta=portb